package jartege ; 

import java.util.* ; 

/**
 * This class allows one manipulate floating values.
 *
 * @author Catherine Oriat
 */

public class FloatingValue {

   private FloatingValue() { } 

   // ------------------------------------------------------------------------
   // Values of type double
   // ------------------------------------------------------------------------

   /**
    * Create a double number given the specified sign, mantissa and 
    * exponent.
    *
    * <pre><jml>
    *
    * requires 
    *    0 <= sign && sign <= 1 ; 
    *
    * requires
    *    0 <= mantissa && mantissa < 1L << 52 ; 
    *
    * requires 
    *    0 <= exponent && exponent < 2047 ;  
    *
    * ensures
    *    sign == 0 ==>
    *       (createDouble(0, 0, exponent) <= \result &&
    *        \result < createDouble(0, 0, exponent + 1)) ;
    *
    * ensures
    *    sign == 1 ==>
    *       (-createDouble(0, 0, exponent + 1) < \result &&
    *        \result <= -createDouble(0, 0, exponent)) ; 
    *    
    * </jml></pre>
    */

   static public double createDouble
         (long sign, long mantissa, long exponent) {
      long bits = (sign << 63) | (exponent << 52) | mantissa ; 
      double d = Double.longBitsToDouble(bits) ; 
      return d ; 
   }

   /**
    * Returns the exponent of a double represented according to the IEEE 754
    * floating point double format bit layout.
    *
    * <pre><jml>
    *
    * ensures 
    *    0 <= \result && \result <= 2047 ; 
    *
    * </jml></pre>
    */
     
   static public long exponent(long bits) {
      return (bits >> 52) & 0x7ffL ; 
   }

   /**
    * Returns the exponent of double f.
    *
    * <pre><jml>
    *
    * ensures
    *    0 <= \result && \result <= 2046 ;
    *
    * ensures
    *    \result == 0 ==> Math.abs(f) < createDouble(0, 0, 1) ; 
    *
    * ensures
    *    \result != 0 ==> createDouble(0, 0, \result) <= Math.abs(f) ; 
    *
    * ensures
    *    \result != 0 && \result != 2046 ==>
    *       Math.abs(f) < createDouble(0, 0, \result + 1) ; 
    *
    * </jml></pre>
    */

   static public long exponent(double f) {
      
      long bits = Double.doubleToLongBits(f) ; 
      long sign = sign(bits) ; 
      long mant = mantissa(bits) ; 
      long exp = exponent(bits) ; 

      double abs = Math.abs(f) ; 

      long exp2 ; 
      if (abs < createDouble(0, 0, 1)) {
         exp2 = 0 ; 
      } else {
         exp2 = (long) Math.floor(Math.log(abs) / Math.log(2)) + 1023 ;
      }

      if (exp != exp2) {
         System.out.println("f = " + f) ; 
         System.out.println("sign = " + sign + ", mantissa = " + mant + ", exp = " + exp) ; 
         System.out.println("exp2 = " + exp2) ; 
      }


      
      return exp2 ; 
      // return exponent(bits) ; 
   }

   /**
    * Returns the sign bit of a double represented according to the IEEE 754
    * floating point double format bit layout.
    * Returns 0 if the double if positive, 1 if the double is negative.
    * 
    * <pre><jml>
    *
    * ensures
    *    0 <= \result && \result <= 1 ;
    *
    * </jml></pre>
    */

   static public long sign(long bits) {
      return -(bits >> 63) ; 
   }

   /**
    * Returns 0 if double f is positive, 1 if double f is negative.
    *
    * <pre><jml>
    *
    * ensures
    *    0 <= \result && \result <= 1 ;
    *
    * </jml></pre>
    */

   static public long sign(double f) {
      long bits = Double.doubleToLongBits(f) ;
      return sign(bits) ; 
   }




   /**
    * Returns the mantissa of a double represented according to the IEEE 754
    * floating point double format bit layout.
    *
    * <pre><jml>
    *
    * ensures
    *    0 <= \result && \result < 1L << 52 ;
    *
    * </jml></pre>
    */

   static public long mantissa(long bits) {
      return bits & 0xfffffffffffffL ; 
   }

   /**
    * Returns the mantissa of a double f.
    *
    * <pre><jml>
    *
    * ensures
    *    0 <= \result && \result < 1L << 52 ;
    *
    * </jml></pre>
    */

   static public long mantissa(double f) {
      long bits = Double.doubleToLongBits(f) ;
      return mantissa(bits) ; 
   }

   /**
    * Returns the least double strictly greater than the specified double. 
    *
    * <pre><jml>
    *
    * requires
    *    f != Double.MAX_VALUE ; 
    *
    * ensures
    *    f < \result && (\forall double x ; f < x ; \result <= x) ; 
    *
    * ensures
    *   previousDouble(\result) == f ; 
    *
    * </jml></pre>
    */

   static public double nextDouble(double f) {
      long bits = Double.doubleToLongBits(f) ; 
      long sign = sign(bits) ; 
      long mant = mantissa(bits) ; 
      long exp = exponent(bits) ; 
      if (sign == 0) {
         if (mant == (1L << 52) - 1) {
            mant = 0 ; 
            exp++ ; 
         } else {
            mant++ ; 
         }
      } else {
         if (mant == 0) {
            if (exp == 0) {
               sign = 0 ; 
               mant = 1 ; 
            } else {
               mant = (1L << 52) - 1 ; 
               exp-- ; 
            }
         } else {
            mant-- ; 
         }
      }
      return createDouble(sign, mant, exp) ; 
   }

   /**
    * Returns the greatest double strictly less than the specified double.
    *
    * <pre><jml>
    *
    * requires
    *    f != -Double.MAX_VALUE ;
    *
    * ensures
    *    \result < f && (\forall double x ; x < f ; x <= \result) ;
    *
    * ensures
    *    nextDouble(\result) == f ; 
    *
    * </jml></pre>
    */

   static public double previousDouble(double f) {
      long bits = Double.doubleToLongBits(f) ;
      long sign = sign(bits) ;
      long mant = mantissa(bits) ;
      long exp = exponent(bits) ;
      if (sign == 0) {
         if (mant == 0) {
            if (exp == 0) {
               sign = 1 ; 
               mant = 1 ; 
            } else {
               mant = (1L << 52) - 1 ; 
               exp-- ; 
            }
         } else {
            mant-- ; 
         }
      } else {
         if (mant == (1L << 52) - 1) {
            mant = 0 ; 
            exp++ ; 
         } else {
            mant++ ; 
         }
      }
      return createDouble(sign, mant, exp) ; 
   }

   // ------------------------------------------------------------------------
   // Values of type float
   // ------------------------------------------------------------------------


   /**
    * Create a float number given the specified sign, mantissa and
    * exponent.
    *
    * <pre><jml>
    *
    * requires 
    *    0 <= sign && sign <= 1 ; 
    *
    * requires
    *    0 <= mantissa && mantissa < 1 << 23 ;
    *
    * requires
    *    0 <= exponent && exponent < 255 ;
    *
    * ensures
    *    sign == 0 ==>
    *       (createFloat(0, 0, exponent) <= \result &&
    *        \result < createFloat(0, 0, exponent + 1)) ; 
    *
    * ensures 
    *    sign == 1 ==>
    *       (-createFloat(0, 0, exponent + 1) < \result &&
    *        \result <= -createFloat(0, 0, exponent)) ; 
    *
    * </jml></pre>
    */
 
   static public float createFloat(int sign, int mantissa, int exponent) {
      int bits = (sign << 31) | (exponent << 23) | mantissa ; 
      float f = Float.intBitsToFloat(bits) ; 
      return f ; 
   }

   /**
    * Returns the exponent of float represented according to the IEEE 754
    * floating point format bit layout.
    *
    * <pre><jml>
    *
    * ensures
    *    0 <= \result && \result <= 255 ;
    *
    * </jml></pre>
    */

   static public int exponent(int bits) {
      int result = (bits >> 23) & 0xff ; 
      return (bits >> 23) & 0xff ; 
   }

   /**
    * Returns the sign  of float represented according to the IEEE 754
    * floating point format bit layout.
    * Returns 0 if the float is positive, 1 if the float is negative.
    *
    * <pre><jml>
    *
    * ensures
    *    0 <= \result && \result <= 1 ; 
    *
    * </jml></pre>
    */

   static public int sign(int bits) {
      return -(bits >> 31) ; 
   }

   /**
    * Returns the mantissa of float represented according to the IEEE 754
    * floating point format bit layout.
    *
    * <pre><jml>
    *
    * ensures
    *    0 <= \result && \result < 1 << 23 ;
    *
    * </jml></pre>
    */

   static public int mantissa(int bits) {
      return bits & 0x7fffff ; 
   }

   /**
    * Returns the least float strictly greater than the specified float. 
    *
    * <pre><jml>
    *
    * requires
    *    f != Float.MAX_VALUE ; 
    *
    * ensures
    *    f < \result && (\forall float x ; f < x ; \result <= x) ; 
    *
    * ensures
    *    previousFloat(\result) == f ; 
    *
    * </jml></pre>
    */

   static public float nextFloat(float f) {
      int bits = Float.floatToIntBits(f) ; 
      int sign = sign(bits) ; 
      int mant = mantissa(bits) ; 
      int exp = exponent(bits) ; 
      if (sign == 0) {
         if (mant == (1 << 23) - 1) {
            mant = 0 ; 
            exp++ ; 
         } else {
            mant++ ; 
         }
      } else {
         if (mant == 0) {
            if (exp == 0) {
               sign = 0 ; 
               mant = 1 ; 
            } else {
               mant = (1 << 23) - 1 ; 
               exp-- ; 
            }
         } else {
            mant-- ; 
         }
      }
      return createFloat(sign, mant, exp) ; 
   }

   /**
    * Returns the greatest float strictly less than the specified float.
    *
    * <pre><jml>
    *
    * requires
    *    f != -Float.MAX_VALUE ;
    *
    * ensures
    *    \result < f && (\forall float x ; x < f ; x <= \result) ;
    *
    * ensures
    *    nextFloat(\result) == f ; 
    *
    * </jml></pre>
    */

   static public float previousFloat(float f) {
      int bits = Float.floatToIntBits(f) ;
      int sign = sign(bits) ;
      int mant = mantissa(bits) ;
      int exp = exponent(bits) ;
      if (sign == 0) {
         if (mant == 0) {
            if (exp == 0) {
               sign = 1 ; 
               mant = 1 ; 
            } else {
               mant = (1 << 23) - 1 ; 
               exp-- ; 
            }
         } else {
            mant-- ; 
         }
      } else {
         if (mant == (1 << 23) - 1) {
            mant = 0 ; 
            exp++ ; 
         } else {
            mant++ ; 
         }
      }
      return createFloat(sign, mant, exp) ; 
   }







}

